/*
 * Copyright 2012-17 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;

import org.openmuc.jdlms.internal.SecSuiteAccessor;
import org.openmuc.jdlms.sessionlayer.SessionLayer;
import org.openmuc.jdlms.settings.client.ConfirmedMode;
import org.openmuc.jdlms.settings.client.ReferencingMethod;
import org.openmuc.jdlms.settings.client.Settings;

/**
 * Class to create a {@link DlmsConnection}.
 * 
 * @param <T>
 *            the concrete connection builder.
 * 
 * @see TcpConnectionBuilder
 * @see SerialConnectionBuilder
 */
public abstract class ConnectionBuilder<T extends ConnectionBuilder<T>> {

    private byte[] systemTitle;

    private int challengeLength;

    private long deviceId;

    private int responseTimeout;
    private final int messageFragmentTimeout;

    private String manufactureId;

    private int logicalDeviceId;

    private int clientId;

    private SecuritySuite securitySuite;

    private int physicalDeviceId;

    private ReferencingMethod referencingMethod;

    private Map<ObisCode, SnObjectInfo> snObjectMapping;

    /**
     * Create a new connection builder, with default settings.
     */
    public ConnectionBuilder() {
        this.systemTitle = new byte[] { 0x4d, 0x4d, 0x4d, 0, 0, 0, 0, 1 };

        this.challengeLength = 64;

        this.deviceId = 1;

        this.responseTimeout = 20000;
        this.messageFragmentTimeout = 5000;
        this.manufactureId = "";

        this.clientId = 16;
        this.logicalDeviceId = 1;

        this.securitySuite = SecuritySuite.builder().build();

        this.physicalDeviceId = 0;

        this.referencingMethod = ReferencingMethod.LOGICAL;

    }

    /**
     * Set the referencing method used to address COSEM objects in the remote meter.
     * 
     * <p>
     * The default referencing method is {@link ReferencingMethod#LOGICAL}.
     * </p>
     * 
     * @param referencingMethod
     *            the referencing method.
     * @return the ConnectionBuilder
     */
    public T setReferencingMethod(ReferencingMethod referencingMethod) {
        this.referencingMethod = referencingMethod;

        return self();
    }

    /**
     * Set the LN -&gt; SN mapping, so the connection does not have to retrieve the mapping when opening the SN
     * connection.
     * 
     * <p>
     * Access the mapping of a short name addressing connection with
     * {@linkplain SnObjectInfo#retrieveLnSnMappingFrom(DlmsConnection)}.
     * </p>
     * 
     * <p>
     * This also sets the referencing method to short.
     * </p>
     * 
     * @param snObjectMapping
     *            the OBIS code to SN object info map
     * @return the ConnectionBuilder
     * @see SnObjectInfo#retrieveLnSnMappingFrom(DlmsConnection)
     */
    public T setSnObjectMapping(Map<ObisCode, SnObjectInfo> snObjectMapping) {
        this.snObjectMapping = new HashMap<>(snObjectMapping);
        setReferencingMethod(ReferencingMethod.SHORT);
        return self();
    }

    /**
     * Set {@link SecuritySuite}.
     * 
     * <p>
     * Default authentication is NONE.
     * </p>
     * 
     * @param securitySuite
     *            authentication object
     * @return the ConnectionBuilder
     */
    public T setSecuritySuite(SecuritySuite securitySuite) {
        this.securitySuite = securitySuite;

        return self();
    }

    /**
     * Set the client ID which should be used to connect to the server.
     * 
     * <p>
     * Default client ID is 16 (Public Client).
     * </p>
     * 
     * @param clientId
     *            the client id.
     * @return the ConnectionBuilder
     */
    public T setClientId(int clientId) {
        this.clientId = clientId;

        return self();
    }

    /**
     * Set the logical device ID of the logical device in the physical server.
     * 
     * <p>
     * Default logical device ID is 1 (Management Logical Device).
     * </p>
     * 
     * @param logicalDeviceId
     *            the logical device ID.
     * @return the ConnectionBuilder
     */
    public T setLogicalDeviceId(int logicalDeviceId) {
        this.logicalDeviceId = logicalDeviceId;

        return self();
    }

    /**
     * Change the used challenge length.
     * 
     * <p>
     * The length must be in the range {@code [8, 64]}. Default is {@code 64}.
     * </p>
     * 
     * @param challengeLength
     *            challenge length
     * @return the ConnectionBuilder
     * 
     * @throws IllegalArgumentException
     *             if the integer is not in the range {@code [8, 64]}.
     */
    public T setChallengeLength(int challengeLength) {
        final int minLength = 8;
        final int maxLength = 64;

        if (challengeLength < minLength || challengeLength > maxLength) {
            throw new IllegalArgumentException("Challenge length has to be between " + minLength + " and " + maxLength);
        }

        this.challengeLength = challengeLength;
        return self();
    }

    /**
     * Set the used device ID.
     * 
     * <p>
     * Not the be mistaken with the logical device ID.
     * </p>
     * 
     * @param deviceId
     *            the device ID
     * @return the ConnectionBuilder
     * 
     */
    public T setDeviceId(long deviceId) {
        this.deviceId = deviceId;
        return self();
    }

    /**
     * Sets the physical device address.
     * 
     * <p>
     * This is only relevant if a connection over HDLC is used. In this case the physical device address is by default
     * 0.
     * </p>
     * 
     * @param physicalDeviceAddress
     *            the physical device address.
     * 
     * @return the ConnectionBuilder
     */
    public T setPhysicalDeviceAddress(int physicalDeviceAddress) {
        this.physicalDeviceId = physicalDeviceAddress;
        return self();
    }

    /**
     * Sets the time in milliseconds the client waits for a response.
     * 
     * @param responseTimeout
     *            time in milliseconds.
     * @return the ConnectionBuilder
     */
    public T setResponseTimeout(int responseTimeout) {
        this.responseTimeout = responseTimeout;

        return self();
    }

    /**
     * Sets the manufacturer ID.
     * 
     * @param manufactureId
     *            the manufacturer ID
     * @return the ConnectionBuilder
     */
    public T setManufactureId(String manufactureId) {
        this.manufactureId = manufactureId;

        return self();
    }

    /**
     * Set the client's system title. It consists of 8 bytes: 3 characters for the manufacturer ID and 5 bytes for the
     * device ID. The default is "MMM" (manufacturer ID) and 1 (device ID).
     * 
     * @param manufacturerId
     *            the manufacturer ID
     * @param deviceId
     *            the device ID
     * @return the ConnectionBuilder
     */
    public T setSystemTitle(String manufacturerId, long deviceId) {
        this.deviceId = deviceId;

        this.systemTitle = new byte[8];

        byte[] manufacturerIdBytes = manufacturerId.getBytes(StandardCharsets.US_ASCII);
        systemTitle[0] = manufacturerIdBytes[0];
        systemTitle[1] = manufacturerIdBytes[1];
        systemTitle[2] = manufacturerIdBytes[2];
        systemTitle[3] = (byte) (deviceId >> 4 * 8);
        systemTitle[4] = (byte) (deviceId >> 3 * 8);
        systemTitle[5] = (byte) (deviceId >> 2 * 8);
        systemTitle[6] = (byte) (deviceId >> 1 * 8);
        systemTitle[7] = (byte) (deviceId);

        return self();
    }

    /**
     * Builds a new DLMS/COSEM Connection.
     * 
     * <p>
     * This method is used if a custom session layer or transport layer is used.
     * </p>
     * 
     * @param settings
     *            the settings for the DLMS/COSEM connection.
     * @param sessionLayer
     *            the session layer.
     * @return A new {@link DlmsConnection} with the given settings.
     * @throws IOException
     *             if an error occurs, while connecting to the meter.
     * 
     * @see #build()
     */
    protected final DlmsConnection buildConnection(Settings settings, SessionLayer sessionLayer) throws IOException {

        DlmsConnection connection;
        if (settings.referencingMethod() == ReferencingMethod.LOGICAL) {
            connection = new DlmsLnConnection(settings, sessionLayer);
        }
        else {
            connection = new DlmsSnConnection(settings, sessionLayer, snObjectMapping);
        }

        connection.connect();

        return connection;
    }

    /**
     * Builds a new DLMS/COSEM Connection, which is connected to the remote meter.
     * 
     * @return A new {@link DlmsConnection} with the given settings.
     * @throws IOException
     *             if an error occurs, while connecting to the meter.
     */
    public abstract DlmsConnection build() throws IOException;

    @SuppressWarnings("unchecked")
    private T self() {
        return (T) this;
    }

    /**
     * This Settings can be overridden by an implementing Settings.
     */
    protected abstract class SettingsImpl implements Settings {
        private final byte[] systemTitle;
        private final int challengeLength;
        private final long deviceId;
        private final int responseTimeout;
        private final int messageFragmentTimeout;
        private final String manufactureId;
        private final int clientAccessPoint;
        private final int logicalDeviceAddress;
        private final SecuritySuite securitySuite;
        private final int physicalDeviceId;
        private final ReferencingMethod referencingMethod;

        public SettingsImpl(ConnectionBuilder<?> builder) {
            this.systemTitle = builder.systemTitle;
            this.challengeLength = builder.challengeLength;
            this.deviceId = builder.deviceId;
            this.responseTimeout = builder.responseTimeout;
            this.messageFragmentTimeout = builder.messageFragmentTimeout;
            this.manufactureId = builder.manufactureId;
            this.clientAccessPoint = builder.clientId;
            this.logicalDeviceAddress = builder.logicalDeviceId;
            this.securitySuite = builder.securitySuite;
            this.physicalDeviceId = builder.physicalDeviceId;
            this.referencingMethod = builder.referencingMethod;
        }

        @Override
        public ReferencingMethod referencingMethod() {
            return this.referencingMethod;
        }

        @Override
        public SecuritySuite securitySuite() {
            synchronized (this.securitySuite) {
                return this.securitySuite;
            }
        }

        @Override
        public int challengeLength() {
            return this.challengeLength;
        }

        @Override
        public byte[] systemTitle() {
            return this.systemTitle;
        }

        @Override
        public long deviceId() {
            return this.deviceId;
        }

        @Override
        public int responseTimeout() {
            return this.responseTimeout;
        }

        @Override
        public int messageFragmentTimeout() {
            return this.messageFragmentTimeout;
        }

        @Override
        public String manufactureId() {
            return this.manufactureId;
        }

        @Override
        public ConfirmedMode confirmedMode() {
            return ConfirmedMode.CONFIRMED;
        }

        @Override
        public int logicalDeviceId() {
            return this.logicalDeviceAddress;
        }

        @Override
        public int physicalDeviceId() {
            return this.physicalDeviceId;
        }

        @Override
        public int clientId() {
            return this.clientAccessPoint;
        }

        @Override
        public void updateAuthenticationKey(byte[] authenticationKey) {
            synchronized (securitySuite) {
                ((SecSuiteAccessor) this.securitySuite).updateAuthentciationKey(authenticationKey);
            }
        }

        @Override
        public void updateGlobalEncryptionKey(byte[] globalEncryptionKey) {
            synchronized (securitySuite) {
                ((SecSuiteAccessor) this.securitySuite).updateGlobalUnicastEncryptionKey(globalEncryptionKey);
            }
        }

    }
}
