/*
 * Copyright 2012-17 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms;

import static org.openmuc.jdlms.internal.DlmsEnumParser.enumValueFrom;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;

import org.openmuc.jdlms.JDlmsException.ExceptionId;
import org.openmuc.jdlms.JDlmsException.Fault;
import org.openmuc.jdlms.SecuritySuite.EncryptionMechanism;
import org.openmuc.jdlms.datatypes.DataObject;
import org.openmuc.jdlms.interfaceclass.method.AssociationSnMethod;
import org.openmuc.jdlms.internal.ConformanceHelper;
import org.openmuc.jdlms.internal.ContextId;
import org.openmuc.jdlms.internal.DataConverter;
import org.openmuc.jdlms.internal.SnInterfaceClass;
import org.openmuc.jdlms.internal.SnInterfaceClassList;
import org.openmuc.jdlms.internal.WellKnownInstanceIds;
import org.openmuc.jdlms.internal.asn1.axdr.AxdrType;
import org.openmuc.jdlms.internal.asn1.axdr.types.AxdrOctetString;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu;
import org.openmuc.jdlms.internal.asn1.cosem.Conformance;
import org.openmuc.jdlms.internal.asn1.cosem.Cosem_Attribute_Descriptor;
import org.openmuc.jdlms.internal.asn1.cosem.Cosem_Date_Time;
import org.openmuc.jdlms.internal.asn1.cosem.Cosem_Object_Instance_Id;
import org.openmuc.jdlms.internal.asn1.cosem.Data;
import org.openmuc.jdlms.internal.asn1.cosem.Data.SubSeqOf_structure;
import org.openmuc.jdlms.internal.asn1.cosem.EVENT_NOTIFICATION_Request;
import org.openmuc.jdlms.internal.asn1.cosem.InformationReportRequest;
import org.openmuc.jdlms.internal.asn1.cosem.Integer16;
import org.openmuc.jdlms.internal.asn1.cosem.Integer8;
import org.openmuc.jdlms.internal.asn1.cosem.ReadRequest;
import org.openmuc.jdlms.internal.asn1.cosem.ReadResponse;
import org.openmuc.jdlms.internal.asn1.cosem.Unsigned16;
import org.openmuc.jdlms.internal.asn1.cosem.Variable_Access_Specification;
import org.openmuc.jdlms.internal.asn1.cosem.WriteRequest;
import org.openmuc.jdlms.internal.asn1.cosem.WriteResponse;
import org.openmuc.jdlms.internal.asn1.cosem.WriteResponse.SubChoice;
import org.openmuc.jdlms.sessionlayer.SessionLayer;
import org.openmuc.jdlms.settings.client.Settings;

/**
 * Variant of the connection class using decrypt messages with short name referencing to communicate with the remote
 * smart meter
 */
class DlmsSnConnection extends DlmsConnection {

    // Allow read/write
    // Allow unconfirmed write
    // Allow information report
    // Allow multiple references
    // Allow parameterized access
    /**
     * Bit field containing all operations this client can perform
     */
    private static Conformance PROPOSED_CONFORMANCE = new Conformance(
            new byte[] { (byte) 0x1C, (byte) 0x03, (byte) 0x20 }, 24);

    /**
     * Short name referring to the list of all accessible Cosem Objects on the smart meter
     */
    private static Integer16 ASSOCIATION_OBJECT_LIST = new Integer16((short) 0xFA08);

    private Map<ObisCode, SnObjectInfo> snObjectMapping;
    private volatile boolean mapIsInitialized;

    private final SnQueue<ReadResponse> readResponseQueue;
    private final SnQueue<WriteResponse> writeResponseQueue;

    DlmsSnConnection(Settings settings, SessionLayer transportLayerCon, Map<ObisCode, SnObjectInfo> snObjectMapping)
            throws IOException {
        super(settings, transportLayerCon);

        this.readResponseQueue = new SnQueue<>();
        this.writeResponseQueue = new SnQueue<>();

        this.mapIsInitialized = false;

        if (snObjectMapping != null && !snObjectMapping.isEmpty()) {
            this.snObjectMapping = snObjectMapping;
        }
        else {
            this.snObjectMapping = new LinkedHashMap<>();
        }
    }

    @Override
    public synchronized List<GetResult> get(boolean highPriority, List<AttributeAddress> params) throws IOException {
        List<Variable_Access_Specification> shortNames;
        try {
            shortNames = getVariableGetList(params);
        } catch (IllegalArgumentException e) {
            ArrayList<GetResult> res = new ArrayList<>(params.size());

            for (int i = 0; i < params.size(); i++) {
                res.add(new GetResult(AccessResultCode.OBJECT_UNAVAILABLE));
            }
            return res;
        }

        ReadRequest request = new ReadRequest();
        for (Variable_Access_Specification name : shortNames) {
            request.add(name);
        }

        COSEMpdu pdu = new COSEMpdu();
        pdu.setreadRequest(request);
        send(pdu);

        ReadResponse response = readResponseQueue.poll();

        List<GetResult> result = new ArrayList<>(response.list().size());
        for (ReadResponse.SubChoice data : response.list()) {
            GetResult resultItem;

            if (data.getChoiceIndex() == ReadResponse.SubChoice.Choices.DATA) {
                DataObject dat = DataConverter.convertDataToDataObject(data.data);
                resultItem = new GetResult(dat);
            }
            else {
                AccessResultCode resultCode = enumValueFrom(data.data_access_error, AccessResultCode.class);
                resultItem = new GetResult(resultCode);
            }

            result.add(resultItem);
        }

        return result;
    }

    @Override
    public synchronized List<AccessResultCode> set(boolean highPriority, List<SetParameter> params) throws IOException {
        List<Variable_Access_Specification> shortNames = getVariableSetList(params);

        return confirmedWrite(params, shortNames);
    }

    private List<AccessResultCode> confirmedWrite(List<SetParameter> params,
            List<Variable_Access_Specification> shortNames) throws IOException {
        List<AccessResultCode> result;
        WriteRequest request = new WriteRequest();
        request.list_of_data = new WriteRequest.SubSeqOf_list_of_data();
        request.variable_access_specification = new WriteRequest.SubSeqOf_variable_access_specification();

        Iterator<Variable_Access_Specification> shortNamesIter = shortNames.iterator();
        Iterator<SetParameter> paramsIter = params.iterator();
        while (paramsIter.hasNext() && shortNamesIter.hasNext()) {
            request.variable_access_specification.add(shortNamesIter.next());
            request.list_of_data.add(DataConverter.convertDataObjectToData(paramsIter.next().data()));
        }

        COSEMpdu pdu = new COSEMpdu();
        pdu.setwriteRequest(request);
        send(pdu);

        WriteResponse responses = writeResponseQueue.poll();

        result = new ArrayList<>(responses.size());
        List<SubChoice> list = responses.list();
        for (WriteResponse.SubChoice response : list) {
            AccessResultCode item;
            if (response.getChoiceIndex() == WriteResponse.SubChoice.Choices.SUCCESS) {
                item = AccessResultCode.SUCCESS;
            }
            else {
                item = enumValueFrom(response.data_access_error, AccessResultCode.class);
            }
            result.add(item);
        }
        return result;
    }

    @Override
    public synchronized List<MethodResult> action(boolean priority, List<MethodParameter> params) throws IOException {
        if (params.size() > 1 && ConformanceHelper.multipleReferenceIsAllowedFor(negotiatedFeatures())) {
            return callEachActionIndividually(priority, params);
        }

        return confirmedMethdoCall(params);
    }

    private List<MethodResult> confirmedMethdoCall(List<MethodParameter> params) throws IOException {
        List<MethodResult> result;
        ReadRequest methodsWithReturn = new ReadRequest();
        WriteRequest methodsWithoutReturn = new WriteRequest();

        /*
         * This list is used to undo the following optimization after all methods have been invoked
         */
        List<Boolean> returnList = new ArrayList<>(params.size());

        /*
         * To optimize network load, split requested methods into methods with return value and methods without
         */
        for (MethodParameter param : params) {

            SnObjectInfo objectInfo = accessSnObjectInfo(param);

            SnInterfaceClass classInfo = SnInterfaceClassList.classInfoFor(objectInfo.getClassId(),
                    objectInfo.getVersion());

            Integer16 variableName = calcVarName(param, classInfo);

            returnList.add(classInfo.hasReturnType(param.getId()));

            if (classInfo.hasReturnType(param.getId())) {
                Variable_Access_Specification access = buildMethodCallWithReturn(param, variableName);

                methodsWithReturn.add(access);
            }
            else {
                Variable_Access_Specification access = buildMehodWithNoReturn(variableName);
                methodsWithoutReturn.variable_access_specification.add(access);
                methodsWithoutReturn.list_of_data
                        .add(DataConverter.convertDataObjectToData(param.getMethodParameter()));
            }
        }

        COSEMpdu pdu = new COSEMpdu();
        WriteResponse responseWithoutReturn = null;
        ReadResponse responseWithReturn = null;

        /*
         * If there are methods with return value requested, send all of them now
         */
        if (!methodsWithoutReturn.variable_access_specification.list().isEmpty()) {
            pdu.setwriteRequest(methodsWithoutReturn);
            send(pdu);

            responseWithoutReturn = this.writeResponseQueue.poll();
        }

        /*
         * If there are methods without return value requested, send all of them now
         */
        if (!methodsWithReturn.list().isEmpty()) {
            pdu.setreadRequest(methodsWithReturn);
            send(pdu);

            responseWithReturn = this.readResponseQueue.poll();
        }

        if (responseWithReturn == null) {
            throw new ResponseTimeoutException("Get timed out..");
        }

        result = new ArrayList<>(params.size());
        int responseWithReturnIndex = 0;
        int responseWithoutReturnIndex = 0;

        // Undo earlier split into methods with and without return value, so
        // the order of result items matches the order of called methods
        for (Boolean withReturn : returnList) {
            MethodResultCode resultCode = null;
            DataObject returnValue = null;
            if (withReturn) {
                if (responseWithReturn.get(responseWithReturnIndex)
                        .getChoiceIndex() == ReadResponse.SubChoice.Choices.DATA) {
                    returnValue = DataConverter
                            .convertDataToDataObject(responseWithReturn.get(responseWithReturnIndex).data);
                    resultCode = MethodResultCode.SUCCESS;
                }
                else {

                    resultCode = enumValueFrom(responseWithReturn.get(responseWithReturnIndex).data_access_error,
                            MethodResultCode.class);

                }
                responseWithReturnIndex++;
            }
            else {
                if (responseWithoutReturn.get(responseWithoutReturnIndex)
                        .getChoiceIndex() == WriteResponse.SubChoice.Choices.DATA_ACCESS_ERROR) {
                    resultCode = enumValueFrom(responseWithoutReturn.get(responseWithoutReturnIndex).data_access_error,
                            MethodResultCode.class);
                }
                else {
                    resultCode = MethodResultCode.SUCCESS;
                }
                responseWithoutReturnIndex++;
            }

            result.add(new MethodResult(resultCode, returnValue));
        }
        return result;
    }

    private Variable_Access_Specification buildMehodWithNoReturn(Integer16 variableName) {
        Variable_Access_Specification access = new Variable_Access_Specification();
        access.setvariable_name(variableName);
        return access;
    }

    private Variable_Access_Specification buildMethodCallWithReturn(MethodParameter param, Integer16 variableName) {
        Variable_Access_Specification access = new Variable_Access_Specification();

        if (param.getMethodParameter() == null) {
            access.setvariable_name(variableName);
        }
        else {
            Variable_Access_Specification.SubSeq_parameterized_access accessParam = new Variable_Access_Specification.SubSeq_parameterized_access(
                    variableName, new Integer8(0), DataConverter.convertDataObjectToData(param.getMethodParameter()));
            access.setparameterized_access(accessParam);
        }
        return access;
    }

    private Integer16 calcVarName(MethodParameter param, SnInterfaceClass classInfo) {
        int variableName = classInfo.firstOffset() + 8 * (param.getId() - 1);
        return new Integer16(variableName);
    }

    private List<MethodResult> callEachActionIndividually(boolean priority, List<MethodParameter> params)
            throws IOException {
        List<MethodResult> result = new ArrayList<>(params.size());
        for (MethodParameter methodParameter : params) {
            MethodResult res = action(priority, methodParameter);
            result.add(res);
        }
        return result;
    }

    @Override
    protected void processPdu(COSEMpdu pdu) {
        try {
            if (pdu.getChoiceIndex() == COSEMpdu.Choices.READRESPONSE) {
                readResponseQueue.put(pdu.readResponse);
            }
            else if (pdu.getChoiceIndex() == COSEMpdu.Choices.WRITERESPONSE) {
                writeResponseQueue.put(pdu.writeResponse);
            }
            else if (pdu.getChoiceIndex() == COSEMpdu.Choices.INFORMATIONREPORTREQUEST) {
                // TODO: implement event listening
                // if (connectionSettings().clientConnectionEventListener() != null) {
                // List<EVENT_NOTIFICATION_Request> eventList = transformEventPdu(pdu.informationReportRequest);
                // for (EVENT_NOTIFICATION_Request event : eventList) {
                // EventNotification notification = DataConverter.toApi(event);
                // connectionSettings().clientConnectionEventListener().onEventReceived(notification);
                // }
                // }
            }
        } catch (InterruptedException e) {
            // TODO
        }
    }

    @Override
    protected Conformance proposedConformance() {
        return PROPOSED_CONFORMANCE;
    }

    private List<Variable_Access_Specification> getVariableSetList(List<SetParameter> params) throws IOException {
        validateListSize(params);

        List<Variable_Access_Specification> result = new ArrayList<>(params.size());
        for (SetParameter setRequestParameter : params) {
            result.add(buildAddressSpec(setRequestParameter.attributeAddress()));
        }
        return result;
    }

    private List<Variable_Access_Specification> getVariableGetList(List<AttributeAddress> params) throws IOException {
        validateListSize(params);

        List<Variable_Access_Specification> result = new ArrayList<>(params.size());
        for (AttributeAddress param : params) {
            result.add(buildAddressSpec(param));
        }
        return result;
    }

    private Variable_Access_Specification buildAddressSpec(AttributeAddress attributeAddress) throws IOException {
        SnObjectInfo info = accessSnObjectInfo(attributeAddress);

        Variable_Access_Specification accessSpec = new Variable_Access_Specification();

        Integer16 variableName = new Integer16(info.getBaseName() + 8 * (attributeAddress.getId() - 1));
        if (attributeAddress.accessSelection() == null) {
            accessSpec.setvariable_name(variableName);
        }
        else {
            if (!ConformanceHelper.parameterizedAccessAllowedFor(negotiatedFeatures())) {
                throw new IllegalArgumentException("Connection doesn't allow access selection");
            }
            accessSpec.setparameterized_access(new Variable_Access_Specification.SubSeq_parameterized_access(
                    variableName, new Integer8(attributeAddress.accessSelection().getAccessSelector()),
                    DataConverter.convertDataObjectToData(attributeAddress.accessSelection().getAccessParameter())));
        }
        return accessSpec;
    }

    private SnObjectInfo accessSnObjectInfo(CosemResourceDescriptor cosemResourceDescriptor) throws IOException {
        ObisCode instanceId = cosemResourceDescriptor.getInstanceId();
        SnObjectInfo info = snObjectMapping.get(instanceId);

        if (info == null && !mapIsInitialized) {

            try {
                SnObjectInfo objectInfo = getVariableInfo(cosemResourceDescriptor);
                snObjectMapping.put(instanceId, objectInfo);
                info = objectInfo;
            } catch (IOException e) {
                initializeLnMap();
                info = snObjectMapping.get(instanceId);
            }
        }

        if (info == null) {
            throw new IllegalArgumentException(
                    "Object " + instanceId + " unknown to the smart meter. Try an other address.");
        }

        return info;
    }

    private void validateListSize(List<?> params) {
        if (params == null || params.isEmpty()) {
            throw new IllegalArgumentException("No parameter provided");
        }
        if (params.size() > 1 && !ConformanceHelper.multipleReferenceIsAllowedFor(negotiatedFeatures())) {
            throw new IllegalArgumentException("Connection does not allow access to multiple parameters in one call");
        }
    }

    private SnObjectInfo getVariableInfo(CosemResourceDescriptor param) throws IOException {
        if (!ConformanceHelper.parameterizedAccessAllowedFor(negotiatedFeatures())) {
            throw new IOException("Connection does not allow parameterized actions");
        }

        ReadRequest request = new ReadRequest();
        Variable_Access_Specification getBaseName = new Variable_Access_Specification();
        Data filter = new Data();
        filter.setstructure(new Data.SubSeqOf_structure());
        filter.structure.add(new Data());
        filter.structure.add(new Data());
        filter.structure.get(0).setlong_unsigned(new Unsigned16(param.getClassId()));
        filter.structure.get(1).setoctet_string(
                new AxdrOctetString(new Cosem_Object_Instance_Id(param.getInstanceId().bytes()).getValue()));
        Variable_Access_Specification.SubSeq_parameterized_access parametrizedAccess = new Variable_Access_Specification.SubSeq_parameterized_access(
                ASSOCIATION_OBJECT_LIST, new Integer8(2), filter);
        getBaseName.setparameterized_access(parametrizedAccess);
        request.add(getBaseName);

        COSEMpdu pdu = new COSEMpdu();
        pdu.setreadRequest(request);
        send(pdu);

        ReadResponse response = this.readResponseQueue.poll();

        if (response.get(0).getChoiceIndex() == ReadResponse.SubChoice.Choices.DATA_ACCESS_ERROR) {
            throw new IOException();
        }

        SnObjectInfo result = new SnObjectInfo(
                (int) response.get(0).data.array.get(0).structure.get(0).long_integer.getValue(),
                (int) response.get(0).data.array.get(0).structure.get(1).long_unsigned.getValue(),
                (int) response.get(0).data.array.get(0).structure.get(2).unsigned.getValue());
        return result;
    }

    Map<ObisCode, SnObjectInfo> getLatestObjectInfoMapping() throws IOException {
        synchronized (this.snObjectMapping) {
            initializeLnMap();
            return new HashMap<>(this.snObjectMapping);
        }
    }

    private void initializeLnMap() throws IOException {
        synchronized (snObjectMapping) {
            if (mapIsInitialized) {
                return;
            }

            ReadRequest request = new ReadRequest();
            Variable_Access_Specification getObjectList = new Variable_Access_Specification();
            getObjectList.setvariable_name(ASSOCIATION_OBJECT_LIST);
            request.add(getObjectList);

            COSEMpdu pdu = new COSEMpdu();
            pdu.setreadRequest(request);
            send(pdu);

            ReadResponse response = this.readResponseQueue.poll();

            if (response.get(0).getChoiceIndex() == ReadResponse.SubChoice.Choices.DATA_ACCESS_ERROR) {
                throw new FatalJDlmsException(ExceptionId.CONNECTION_ESTABLISH_ERROR, Fault.SYSTEM,
                        "Could not access the mapping list.");
            }

            List<Data> resultData = response.get(0).data.array.list();
            for (Data object : resultData) {
                SubSeqOf_structure objectStructur = object.structure;
                int baseName = (int) objectStructur.get(0).long_integer.getValue();
                int classId = (int) objectStructur.get(1).long_unsigned.getValue();
                int version = (int) objectStructur.get(2).unsigned.getValue();

                SnObjectInfo value = new SnObjectInfo(baseName, classId, version);

                byte[] logicalNameOctetStr = objectStructur.get(3).octet_string.getValue();
                ObisCode key = new ObisCode(logicalNameOctetStr);

                snObjectMapping.put(key, value);
            }

            this.mapIsInitialized = true;
        }
    }

    // TODO: see event linstening
    @SuppressWarnings("unused")
    private List<EVENT_NOTIFICATION_Request> transformEventPdu(InformationReportRequest event) throws IOException {
        Cosem_Date_Time convertedTime = null;
        ByteArrayOutputStream oStream = new ByteArrayOutputStream(12);
        if (event.current_time.isUsed()) {
            InputStream iStream = new ByteArrayInputStream(event.current_time.getValue().getValue());
            int buffer = iStream.read() - 0x30;

            // Converting year
            buffer = buffer * 10 + iStream.read() - 0x30;
            buffer = buffer * 10 + iStream.read() - 0x30;
            buffer = buffer * 10 + iStream.read() - 0x30;
            oStream.write((byte) ((buffer >> 8) & 0xFF));
            oStream.write((byte) (buffer & 0xFF));

            // Month
            buffer = iStream.read();
            buffer = buffer * 10 + iStream.read();
            oStream.write((byte) (buffer & 0xFF));

            // Day
            buffer = iStream.read();
            buffer = buffer * 10 + iStream.read();
            oStream.write((byte) (buffer & 0xFF));

            // Day of week not specified
            oStream.write((byte) 0xFF);

            // Hour
            buffer = iStream.read();
            buffer = buffer * 10 + iStream.read();
            oStream.write((byte) (buffer & 0xFF));

            // Minute
            buffer = iStream.read();
            buffer = buffer * 10 + iStream.read();
            oStream.write((byte) (buffer & 0xFF));

            // Second
            buffer = iStream.read();
            buffer = buffer * 10 + iStream.read();
            oStream.write((byte) (buffer & 0xFF));

            // Milliseconds specified
            oStream.write((byte) 0xFF);

            // Deviation specified
            oStream.write((byte) 0x80);
            oStream.write((byte) 0x00);

            // Status no error
            oStream.write((byte) 0x00);

            convertedTime = new Cosem_Date_Time(oStream.toByteArray());
        }

        Iterator<Variable_Access_Specification> eventIter = event.variable_access_specification.iterator();
        Iterator<Data> dataIter = event.list_of_data.iterator();

        List<EVENT_NOTIFICATION_Request> result = new ArrayList<>(event.variable_access_specification.size());

        while (eventIter.hasNext() && dataIter.hasNext()) {
            Variable_Access_Specification eventInfo = eventIter.next();
            Data eventData = dataIter.next();

            EVENT_NOTIFICATION_Request notificationRequest = extracted(event, convertedTime, eventInfo, eventData);
            result.add(notificationRequest);
        }

        return result;
    }

    private EVENT_NOTIFICATION_Request extracted(InformationReportRequest event, Cosem_Date_Time convertedTime,
            Variable_Access_Specification eventInfo, Data eventData) {

        Iterator<Map.Entry<ObisCode, SnObjectInfo>> iterator = snObjectMapping.entrySet().iterator();

        while (iterator.hasNext()) {
            ObisCode shortNameKey = iterator.next().getKey();
            SnObjectInfo shortNameInfo = snObjectMapping.get(shortNameKey);

            if (SnInterfaceClassList.classInfoFor(shortNameInfo.getClassId(), shortNameInfo.getVersion())
                    .isInRange((int) eventInfo.variable_name.getValue(), shortNameInfo.getBaseName())) {

                Cosem_Attribute_Descriptor logicalNameInfo = new Cosem_Attribute_Descriptor(
                        new Unsigned16(shortNameInfo.getClassId()), new Cosem_Object_Instance_Id(shortNameKey.bytes()),
                        new Integer8((eventInfo.variable_name.getValue() - shortNameInfo.getBaseName()) / 8 + 1));

                EVENT_NOTIFICATION_Request listItem = new EVENT_NOTIFICATION_Request();
                listItem.cosem_attribute_descriptor = logicalNameInfo;
                listItem.attribute_value = eventData;

                if (event.current_time.isUsed()) {
                    listItem.time.setValue(convertedTime);
                }
                return listItem;
            }
        }
        return null;
    }

    @Override
    protected MethodResult hlsAuthentication(byte[] processedChallenge) throws IOException {
        DataObject param = DataObject.newOctetStringData(processedChallenge);

        WellKnownInstanceIds.SECURITY_SETUP_ID.length();
        MethodParameter authenticate = new MethodParameter(AssociationSnMethod.REPLY_TO_HLS_AUTHENTICATION,
                new ObisCode(0, 0, 40, 0, 0, 255), param); // should be association sn class?

        return action(false, authenticate);
    }

    @Override
    protected void validateReferencingMethod() throws IOException {
        // If the first byte of the Conformance bit string is 0, then neither
        // read nor write are allowed, a sign that this smart meter cannot
        // communicate with LN referencing.
        if (negotiatedFeatures().value[0] == 0) {
            disconnect();
            throw new IOException("Wrong referencing method. Remote smart meter +" + "can't use SN referencing");
        }
    }

    @Override
    protected ContextId getContextId() {
        if (connectionSettings().securitySuite().getEncryptionMechanism() != EncryptionMechanism.NONE) {
            return ContextId.SHORT_NAME_REFERENCING_WITH_CIPHERING;
        }
        else {
            return ContextId.SHORT_NAME_REFERENCING_NO_CIPHERING;
        }
    }

    private class SnQueue<E extends AxdrType> {
        private final BlockingQueue<E> q;

        public SnQueue() {
            this.q = new ArrayBlockingQueue<>(3);
        }

        public void put(E e) throws InterruptedException {
            this.q.put(e);
        }

        public E poll() throws IOException {
            try {
                E result = q.poll(connectionSettings().responseTimeout(), TimeUnit.MILLISECONDS);

                if (result == null) {
                    throw new ResponseTimeoutException("Request timed out..");
                }

                return result;
            } catch (InterruptedException e) {
                // should not occur
                throw new IOException("Interrupted while waiting for incoming response", e);
            }
        }

    }

}
