/*
 * Copyright 2012-17 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.internal.sessionlayer.hdlc;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.text.MessageFormat;

public class HdlcFrame {
    private static byte FLAG = 0x7E;

    private final FrameType frameType;

    private byte[] informationField;

    private int sendSequence;
    private int receiveSequence;
    private boolean segmented;

    private byte controlField;

    private final HdlcAddressPair addressPair;

    public static HdlcFrame decode(byte[] frame) throws IOException, FrameInvalidException {
        final FcsCalc fcsCalc = new FcsCalc();
        ByteBuffer buffer = ByteBuffer.wrap(frame);

        byte frameFormatH = buffer.get();
        if ((frameFormatH & 0xF0) != 0xA0) {
            throw new FrameInvalidException("Illegal frameformat");
        }
        boolean segmented = (0x08 & frameFormatH) == 0x08;

        fcsCalc.update(frameFormatH);
        // content can be ignored, has been read in MessageDecoder
        fcsCalc.update(buffer.get());

        HdlcAddress destination = readAddress(fcsCalc, buffer);
        HdlcAddress source = readAddress(fcsCalc, buffer);

        byte controlField = buffer.get();
        fcsCalc.update(controlField);

        FrameType frameType = FrameType.frameTypeFor(controlField & 0xFF);
        if (frameType == FrameType.ERR_INVALID_TYPE) {
            FrameRejectReason reason = new FrameRejectReason(controlField);

            throw new FrameInvalidException(MessageFormat.format("Control field unknown {0}", controlField), reason);
        }
        fcsCalc.update(buffer.get());
        fcsCalc.update(buffer.get());
        fcsCalc.validateCurrentFcsValue();

        int sendSequence = 0;
        int receiveSequence = 0;

        if (frameType == FrameType.INFORMATION) {
            // Send sequence number are the bits 1 to 3 of the frame type
            // field
            sendSequence = (controlField & 0x0E) >> 1;
        }
        if (frameType == FrameType.INFORMATION || frameType == FrameType.RECEIVE_READY
                || frameType == FrameType.RECEIVE_NOT_READY) {
            // Receive sequence number are the bits 5 to 7 of the frame type
            // field
            receiveSequence = (controlField & 0xE0) >> 5;
        }
        byte[] informationField = new byte[0];
        if (buffer.hasRemaining()) {
            int infoLength = buffer.remaining() - 2;
            informationField = new byte[infoLength];

            for (int i = 0; i < infoLength; i++) {
                byte data = buffer.get();
                fcsCalc.update(data);
                informationField[i] = data;
            }
            fcsCalc.update(buffer.get());
            fcsCalc.update(buffer.get());
            fcsCalc.validateCurrentFcsValue();
        }
        HdlcAddressPair addressPair = new HdlcAddressPair(source, destination);
        return new HdlcFrame(frameType, informationField, sendSequence, receiveSequence, segmented, controlField,
                addressPair);
    }

    private static HdlcAddress readAddress(final FcsCalc fcsCalc, ByteBuffer buffer) throws FrameInvalidException {
        byte[] data = new byte[4];
        byte currentByte = 0;
        int length = 0;
        while ((currentByte & 0x01) == 0) {
            if (length == 4) {
                throw new FrameInvalidException("HLDC address is illegal in frame.");
            }
            currentByte = buffer.get();
            fcsCalc.update(currentByte);
            data[length++] = currentByte;
        }

        return HdlcAddress.decode(data, length);
    }

    private HdlcFrame(FrameType frameType, byte[] informationField, int sendSequence, int receiveSequence,
            boolean segmented, byte controlField, HdlcAddressPair addressPair) {
        this.frameType = frameType;
        this.informationField = informationField;
        this.sendSequence = sendSequence;
        this.receiveSequence = receiveSequence;
        this.segmented = segmented;
        this.controlField = controlField;
        this.addressPair = addressPair;
    }

    private HdlcFrame(HdlcAddressPair addressPair, FrameType frameType) {
        this.segmented = false;

        this.sendSequence = -1;
        this.receiveSequence = -1;

        this.frameType = frameType;
        this.addressPair = addressPair;
    }

    public static HdlcFrame newInformationFrame(HdlcAddressPair addressPair, int sendSequence, int receiveSequence,
            byte[] data, boolean segmented) {
        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.INFORMATION);

        hdlcFrame.sendSequence = sendSequence;
        hdlcFrame.receiveSequence = receiveSequence;
        hdlcFrame.informationField = data;
        hdlcFrame.segmented = segmented;

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        hdlcFrame.controlField |= ((sendSequence % 8) << 1);
        hdlcFrame.controlField |= ((receiveSequence % 8) << 5);
        if (!segmented) {
            hdlcFrame.controlField |= 0x10;
        }

        return hdlcFrame;
    }

    public static HdlcFrame newReceiveReadyFrame(HdlcAddressPair addressPair, int receiveSeq, boolean poll) {
        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.RECEIVE_READY);
        hdlcFrame.informationField = null;
        hdlcFrame.segmented = false;

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        hdlcFrame.controlField |= ((receiveSeq % 8) << 5);
        if (poll) {
            hdlcFrame.controlField |= 0x10;
        }

        return hdlcFrame;
    }

    public static HdlcFrame newSetNormalResponseModeFrame(HdlcAddressPair addressPair,
            HdlcParameterNegotiation negotiationParams, boolean poll) throws IOException {

        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.SET_NORMAL_RESPONSEMODE);
        hdlcFrame.informationField = negotiationParams.encode();
        hdlcFrame.segmented = false;

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        if (poll) {
            hdlcFrame.controlField |= 0x10;
        }
        return hdlcFrame;
    }

    public static HdlcFrame newUnnumberedInformationFrame(HdlcAddressPair addressPair, byte[] information,
            boolean poll) {
        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.UNNUMBERED_INFORMATION);

        hdlcFrame.informationField = information;
        hdlcFrame.segmented = false;

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        if (poll) {
            hdlcFrame.controlField |= 0x10;
        }

        return hdlcFrame;
    }

    public static HdlcFrame newDisconnectFrame(HdlcAddressPair addressPair, boolean poll) {
        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.DISCONNECT);

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        if (poll) {
            hdlcFrame.controlField |= 0x10;
        }

        return hdlcFrame;
    }

    // TODO: chick if this is needed??
    public static HdlcFrame newReceiveNotReadyFrame(HdlcAddressPair addressPair, int receiveSeq, boolean poll) {
        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.RECEIVE_NOT_READY);
        hdlcFrame.informationField = null;
        hdlcFrame.segmented = false;

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        hdlcFrame.controlField |= ((receiveSeq % 8) << 5);
        if (poll) {
            hdlcFrame.controlField |= 0x10;
        }

        return hdlcFrame;
    }

    // TODO: chick if this is needed??
    public static HdlcFrame newUnnumberedAcknowledgeFrame(HdlcAddressPair addressPair,
            HdlcParameterNegotiation negotiationParams, boolean poll) throws IOException {
        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.UNNUMBERED_ACKNOWLEDGE);

        if (negotiationParams != null) {
            hdlcFrame.informationField = negotiationParams.encode();
        }
        else {
            hdlcFrame.informationField = new byte[0];
        }
        hdlcFrame.segmented = false;

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        if (poll) {
            hdlcFrame.controlField |= 0x10;
        }

        return hdlcFrame;
    }

    // TODO: chick if this is needed??
    public static HdlcFrame newDisconnectModeFrame(HdlcAddressPair addressPair, byte[] information, boolean poll) {
        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.DISCONNECT_MODE);
        hdlcFrame.informationField = information;
        hdlcFrame.segmented = false;

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        if (poll) {
            hdlcFrame.controlField |= 0x10;
        }

        return hdlcFrame;
    }

    // TODO: chick if this is needed??
    public static HdlcFrame newFrameRejectFrame(HdlcAddressPair addressPair, FrameRejectReason reason, boolean poll) {
        HdlcFrame hdlcFrame = new HdlcFrame(addressPair, FrameType.FRAME_REJECT);
        hdlcFrame.informationField = reason.encode();
        hdlcFrame.segmented = false;

        hdlcFrame.controlField = (byte) hdlcFrame.frameType.value();
        if (poll) {
            hdlcFrame.controlField |= 0x10;
        }

        return hdlcFrame;
    }

    public HdlcAddress getDestinationAddress() {
        return addressPair.destination();
    }

    public HdlcAddress getSourceAddress() {
        return addressPair.source();
    }

    public HdlcAddressPair getAddressPair() {
        return addressPair;
    }

    public FrameType getFrameType() {
        return frameType;
    }

    public byte[] getInformationField() {
        return informationField;
    }

    public int getSendSequence() {
        return sendSequence;
    }

    public int getReceiveSequence() {
        return receiveSequence;
    }

    public boolean isSegmented() {
        return segmented;
    }

    public byte[] encode() throws FrameInvalidException {
        byte[] data = encodeWithoutFlags();
        return ByteBuffer.allocate(data.length + 2).put(FLAG).put(data).put(FLAG).array();
    }

    private byte[] encodeWithoutFlags() throws FrameInvalidException {
        if (frameType == FrameType.ERR_INVALID_TYPE) {
            throw new FrameInvalidException("Frame not initialized prior to encode");
        }

        int length = 2 + getDestinationAddress().length() + getSourceAddress().length() + 1 + 2;
        if (containsInformation()) {
            length += informationField.length + 2;
        }

        ByteBuffer code = ByteBuffer.allocate(length);

        short frameFormat = (short) (0xA000 | length);
        if (segmented) {
            frameFormat |= 0x0800;
        }

        code.putShort(frameFormat);
        code.put(getDestinationAddress().encode());
        code.put(getSourceAddress().encode());
        code.put(controlField);

        FcsCalc fcsCalc = new FcsCalc();
        fcsCalc.update(code.array(), code.position());
        code.put(fcsCalc.fcsValueInBytes());

        if (containsInformation()) {
            fcsCalc.update(fcsCalc.fcsValueInBytes());
            code.put(informationField);
            fcsCalc.update(informationField);
            code.put(fcsCalc.fcsValueInBytes());
        }

        return code.array();
    }

    private boolean containsInformation() {
        return informationField != null;
    }
}
