package org.openmuc.jdlms.sample.server;

import java.io.IOException;
import java.net.InetAddress;

import org.openmuc.jdlms.AccessResultCode;
import org.openmuc.jdlms.AttributeAddress;
import org.openmuc.jdlms.AuthenticationMechanism;
import org.openmuc.jdlms.DlmsAccessException;
import org.openmuc.jdlms.DlmsConnection;
import org.openmuc.jdlms.DlmsInterceptor;
import org.openmuc.jdlms.DlmsInvocationContext;
import org.openmuc.jdlms.DlmsServer;
import org.openmuc.jdlms.GetResult;
import org.openmuc.jdlms.HexConverter;
import org.openmuc.jdlms.LogicalDevice;
import org.openmuc.jdlms.MethodParameter;
import org.openmuc.jdlms.MethodResult;
import org.openmuc.jdlms.SecuritySuite;
import org.openmuc.jdlms.SecuritySuite.EncryptionMechanism;
import org.openmuc.jdlms.SecurityUtils;
import org.openmuc.jdlms.SecurityUtils.KeyId;
import org.openmuc.jdlms.TcpConnectionBuilder;
import org.openmuc.jdlms.datatypes.DataObject;
import org.openmuc.jdlms.interfaceclass.attribute.AssociationLnAttribute;

public class SampleClientServer {
    private final static byte[] AUTHENTICATION_KEY = HexConverter
            .fromShortHexString("5468697349734150617373776f726431");

    private final static byte[] GLOBAL_ENCRYPTION_KEY = HexConverter
            .fromShortHexString("000102030405060708090a0b0c0d0e0f");
    private static final byte[] MASTER_KEY = HexConverter.fromShortHexString("aa0102030405060738090a0b0c0d0eff");;

    private final static int PORT = 6789;
    private final static String MANUFACTURE_ID = "ISE";
    private final static long DEVICE_ID = 9999L;
    private final static String LOGICAL_DEVICE_ID = "L_D_I";

    private final static SecuritySuite AUTHENTICATION_C = SecuritySuite.builder()
            .setAuthenticationKey(AUTHENTICATION_KEY)
            .setGlobalUnicastEncryptionKey(GLOBAL_ENCRYPTION_KEY)
            .setAuthenticationMechanism(AuthenticationMechanism.HLS5_GMAC)
            .setEncryptionMechanism(EncryptionMechanism.AES_GMC_128)
            .build();

    private final static SecuritySuite AUTHENTICATION_S = AUTHENTICATION_C.clone();

    public static void main(String[] args) throws IOException {
        DlmsServer serverConnection;

        try {
            printServer("starting");
            LogicalDevice logicalDevice = new LogicalDevice(1, LOGICAL_DEVICE_ID, "HMM", DEVICE_ID);
            logicalDevice.setMasterKey(MASTER_KEY);
            logicalDevice.addRestriction(16, AUTHENTICATION_S);
            logicalDevice.registerCosemObject(new SampleClass(new DlmsInterceptor() {

                @Override
                public DataObject intercept(DlmsInvocationContext ctx) throws DlmsAccessException {
                    System.out.println("--------------------------------------");
                    System.out.println(ctx.getXDlmsServiceType());
                    System.out.println(ctx.getCosemResourceDescriptor().getClassId());
                    System.out.println(ctx.getCosemResourceDescriptor().getInstanceId());
                    System.out.println(ctx.getCosemResourceDescriptor().getId());

                    System.out.println(ctx.getSecurityPolicy());

                    System.out.println("--------------------------------------");

                    return ctx.proceed();
                }
            }));

            serverConnection = DlmsServer.tcpServerBuilder(PORT).registerLogicalDevice(logicalDevice).build();

            printServer("started");
        } catch (IOException e) {
            throw new IOException("DemoServer: " + e);
        }

        try {
            runDemoClient();

        } finally {
            if (serverConnection != null) {
                serverConnection.close();
            }
            printServer("closed");
        }

    }

    static void runDemoClient() throws IOException {
        TcpConnectionBuilder connectionBuilder = new TcpConnectionBuilder(InetAddress.getLocalHost())
                .setLogicalDeviceId(1)
                .setTcpPort(PORT)
                .setSecuritySuite(AUTHENTICATION_C)
                .setSystemTitle(MANUFACTURE_ID, DEVICE_ID);

        printClient("connecting to server");

        byte[] newKey = GLOBAL_ENCRYPTION_KEY.clone();
        DlmsConnection client = connectionBuilder.build();
        printClient("connected");

        for (int i = 0; i < newKey.length; i++) {
            newKey[i] = (byte) (newKey[i] ^ 42);
        }

        System.out.println("new enckey " + HexConverter.toHexString(newKey));
        MethodParameter keyChangeMethodParam = SecurityUtils.keyChangeMethodParamFor(MASTER_KEY, newKey,
                KeyId.GLOBAL_UNICAST_ENCRYPTION_KEY);

        MethodResult methodResult = client.action(keyChangeMethodParam);
        System.out.println("Change key: " + methodResult.getResultCode());

        client.changeClientGlobalEncryptionKey(newKey);
        final String associationLnInstancId = "0.0.40.0.0.255";

        AccessResultCode resultCode = client
                .get(new AttributeAddress(AssociationLnAttribute.OBJECT_LIST, associationLnInstancId)).getResultCode();
        System.out.println("Req after change key: " + resultCode);

        client.disconnect();

        SecuritySuite newAuth = SecuritySuite.builder()
                .setAuthenticationKey(AUTHENTICATION_KEY)
                .setGlobalUnicastEncryptionKey(newKey)
                .setAuthenticationMechanism(AuthenticationMechanism.HLS5_GMAC)
                .setEncryptionMechanism(EncryptionMechanism.AES_GMC_128)
                .build();
        connectionBuilder.setSecuritySuite(newAuth);
        try (DlmsConnection con = connectionBuilder.build()) {
            MethodResult result = con.action(new MethodParameter(99, "0.0.0.2.1.255", 1));
            System.out.println(result.getResultCode());

            GetResult getResult = con.get(new AttributeAddress(99, "0.0.0.2.1.255", 2));
            System.out.println("--------------------------------");
            System.out.println(getResult.getResultData());
            System.out.println("--------------------------------");

        }

        printClient("closed");
    }

    private static void printClient(String message) {
        System.out.println("DemoClient: " + message);
    }

    private static void printServer(String message) {
        System.out.println("DemoServer: " + message);
    }
}
